@@ -1,28 +1,34 @@
+"""
+information schema implementation.
+
+This module is deprecated and will not be present in this form in SQLAlchemy 0.6.
+
+"""
+from sqlalchemy import util
+
+util.warn_deprecated("the information_schema module is deprecated.")
+
 import sqlalchemy.sql as sql
-import sqlalchemy.engine as engine
-import sqlalchemy.schema as schema
-import sqlalchemy.ansisql as ansisql
-import sqlalchemy.types as sqltypes
-import sqlalchemy.exceptions as exceptions
-from sqlalchemy import *
-from sqlalchemy.ansisql import *
+import sqlalchemy.exc as exc
+from sqlalchemy import select, MetaData, Table, Column, String, Integer
+from sqlalchemy.schema import DefaultClause, ForeignKeyConstraint
 
 ischema = MetaData()
 
-schemata = schema.Table("schemata", ischema,
+schemata = Table("schemata", ischema,
     Column("catalog_name", String),
     Column("schema_name", String),
     Column("schema_owner", String),
     schema="information_schema")
 
-tables = schema.Table("tables", ischema,
+tables = Table("tables", ischema,
     Column("table_catalog", String),
     Column("table_schema", String),
     Column("table_name", String),
     Column("table_type", String),
     schema="information_schema")
 
-columns = schema.Table("columns", ischema,
+columns = Table("columns", ischema,
     Column("table_schema", String),
     Column("table_name", String),
     Column("column_name", String),
@@ -33,23 +39,24 @@ columns = schema.Table("columns", ischema,
     Column("numeric_precision", Integer),
     Column("numeric_scale", Integer),
     Column("column_default", Integer),
+    Column("collation_name", String),
     schema="information_schema")
-    
-constraints = schema.Table("table_constraints", ischema,
+
+constraints = Table("table_constraints", ischema,
     Column("table_schema", String),
     Column("table_name", String),
     Column("constraint_name", String),
     Column("constraint_type", String),
     schema="information_schema")
 
-column_constraints = schema.Table("constraint_column_usage", ischema,
+column_constraints = Table("constraint_column_usage", ischema,
     Column("table_schema", String),
     Column("table_name", String),
     Column("column_name", String),
     Column("constraint_name", String),
     schema="information_schema")
 
-pg_key_constraints = schema.Table("key_column_usage", ischema,
+pg_key_constraints = Table("key_column_usage", ischema,
     Column("table_schema", String),
     Column("table_name", String),
     Column("column_name", String),
@@ -57,7 +64,7 @@ pg_key_constraints = schema.Table("key_column_usage", ischema,
     Column("ordinal_position", Integer),
     schema="information_schema")
 
-#mysql_key_constraints = schema.Table("key_column_usage", ischema,
+#mysql_key_constraints = Table("key_column_usage", ischema,
 #    Column("table_schema", String),
 #    Column("table_name", String),
 #    Column("column_name", String),
@@ -69,7 +76,7 @@ pg_key_constraints = schema.Table("key_column_usage", ischema,
 
 key_constraints = pg_key_constraints
 
-ref_constraints = schema.Table("referential_constraints", ischema,
+ref_constraints = Table("referential_constraints", ischema,
     Column("constraint_catalog", String),
     Column("constraint_schema", String),
     Column("constraint_name", String),
@@ -80,42 +87,26 @@ ref_constraints = schema.Table("referential_constraints", ischema,
     Column("update_rule", String),
     Column("delete_rule", String),
     schema="information_schema")
-                                   
-class ISchema(object):
-    def __init__(self, engine):
-        self.engine = engine
-        self.cache = {}
-
-    def __getattr__(self, name):
-        if name not in self.cache:
-            # This is a bit of a hack.
-            # It would probably be better to have a dict
-            # with just the information_schema tables at
-            # the module level, so as to avoid returning
-            # unrelated objects that happen to be named
-            # 'gen_*'
-            try:
-                gen_tbl = globals()['gen_'+name]
-            except KeyError:
-                raise ArgumentError('information_schema table %s not found' % name)
-            self.cache[name] = gen_tbl.toengine(self.engine)
-        return self.cache[name]
 
 
-def reflecttable(connection, table, ischema_names):
-    
+def table_names(connection, schema):
+    s = select([tables.c.table_name], tables.c.table_schema==schema)
+    return [row[0] for row in connection.execute(s)]
+
+
+def reflecttable(connection, table, include_columns, ischema_names):
     key_constraints = pg_key_constraints
-        
+
     if table.schema is not None:
         current_schema = table.schema
     else:
         current_schema = connection.default_schema_name()
-    
-    s = select([columns], 
+
+    s = select([columns],
         sql.and_(columns.c.table_name==table.name,
         columns.c.table_schema==current_schema),
         order_by=[columns.c.ordinal_position])
-        
+
     c = connection.execute(s)
     found_table = False
     while True:
@@ -126,14 +117,16 @@ def reflecttable(connection, table, ischema_names):
  #       continue
         found_table = True
         (name, type, nullable, charlen, numericprec, numericscale, default) = (
-            row[columns.c.column_name], 
-            row[columns.c.data_type], 
-            row[columns.c.is_nullable] == 'YES', 
+            row[columns.c.column_name],
+            row[columns.c.data_type],
+            row[columns.c.is_nullable] == 'YES',
             row[columns.c.character_maximum_length],
             row[columns.c.numeric_precision],
             row[columns.c.numeric_scale],
             row[columns.c.column_default]
             )
+        if include_columns and name not in include_columns:
+            continue
 
         args = []
         for a in (charlen, numericprec, numericscale):
@@ -142,13 +135,13 @@ def reflecttable(connection, table, ischema_names):
         coltype = ischema_names[type]
         #print "coltype " + repr(coltype) + " args " +  repr(args)
         coltype = coltype(*args)
-        colargs= []
+        colargs = []
         if default is not None:
-            colargs.append(PassiveDefault(sql.text(default)))
-        table.append_column(schema.Column(name, coltype, nullable=nullable, *colargs))
-    
+            colargs.append(DefaultClause(sql.text(default)))
+        table.append_column(Column(name, coltype, nullable=nullable, *colargs))
+
     if not found_table:
-        raise exceptions.NoSuchTableError(table.name)
+        raise exc.NoSuchTableError(table.name)
 
     # we are relying on the natural ordering of the constraint_column_usage table to return the referenced columns
     # in an order that corresponds to the ordinal_position in the key_constraints table, otherwise composite foreign keys
@@ -174,28 +167,27 @@ def reflecttable(connection, table, ischema_names):
             row[colmap[5]],
             row[colmap[6]]
         )
-        #print "type %s on column %s to remote %s.%s.%s" % (type, constrained_column, referred_schema, referred_table, referred_column) 
-        if type=='PRIMARY KEY':
+        #print "type %s on column %s to remote %s.%s.%s" % (type, constrained_column, referred_schema, referred_table, referred_column)
+        if type == 'PRIMARY KEY':
             table.primary_key.add(table.c[constrained_column])
-        elif type=='FOREIGN KEY':
+        elif type == 'FOREIGN KEY':
             try:
                 fk = fks[constraint_name]
             except KeyError:
-                fk = ([],[])
+                fk = ([], [])
                 fks[constraint_name] = fk
             if current_schema == referred_schema:
                 referred_schema = table.schema
             if referred_schema is not None:
-                schema.Table(referred_table, table.metadata, autoload=True, schema=referred_schema, autoload_with=connection)
+                Table(referred_table, table.metadata, autoload=True, schema=referred_schema, autoload_with=connection)
                 refspec = ".".join([referred_schema, referred_table, referred_column])
             else:
-                schema.Table(referred_table, table.metadata, autoload=True, autoload_with=connection)
+                Table(referred_table, table.metadata, autoload=True, autoload_with=connection)
                 refspec = ".".join([referred_table, referred_column])
             if constrained_column not in fk[0]:
                 fk[0].append(constrained_column)
             if refspec not in fk[1]:
                 fk[1].append(refspec)
-    
+
     for name, value in fks.iteritems():
-        table.append_constraint(ForeignKeyConstraint(value[0], value[1], name=name))    
-            
+        table.append_constraint(ForeignKeyConstraint(value[0], value[1], name=name))
