@@ -1,27 +1,97 @@
 # postgres.py
-# Copyright (C) 2005, 2006, 2007 Michael Bayer mike_mp@zzzcomputing.com
+# Copyright (C) 2005, 2006, 2007, 2008, 2009 Michael Bayer mike_mp@zzzcomputing.com
 #
 # This module is part of SQLAlchemy and is released under
 # the MIT License: http://www.opensource.org/licenses/mit-license.php
 
 """Support for the PostgreSQL database.
 
-PostgreSQL supports partial indexes. To create them pass a posgres_where
+Driver
+------
+
+The psycopg2 driver is supported, available at http://pypi.python.org/pypi/psycopg2/ .
+The dialect has several behaviors  which are specifically tailored towards compatibility 
+with this module.
+
+Note that psycopg1 is **not** supported.
+
+Connecting
+----------
+
+URLs are of the form `postgres://user:password@host:port/dbname[?key=value&key=value...]`.
+
+Postgres-specific keyword arguments which are accepted by :func:`~sqlalchemy.create_engine()` are:
+
+* *server_side_cursors* - Enable the usage of "server side cursors" for SQL statements which support
+  this feature.  What this essentially means from a psycopg2 point of view is that the cursor is 
+  created using a name, e.g. `connection.cursor('some name')`, which has the effect that result rows
+  are not immediately pre-fetched and buffered after statement execution, but are instead left 
+  on the server and only retrieved as needed.    SQLAlchemy's :class:`~sqlalchemy.engine.base.ResultProxy`
+  uses special row-buffering behavior when this feature is enabled, such that groups of 100 rows 
+  at a time are fetched over the wire to reduce conversational overhead.
+
+Sequences/SERIAL
+----------------
+
+Postgres supports sequences, and SQLAlchemy uses these as the default means of creating
+new primary key values for integer-based primary key columns.   When creating tables, 
+SQLAlchemy will issue the ``SERIAL`` datatype for integer-based primary key columns, 
+which generates a sequence corresponding to the column and associated with it based on
+a naming convention.
+
+To specify a specific named sequence to be used for primary key generation, use the
+:func:`~sqlalchemy.schema.Sequence` construct::
+
+    Table('sometable', metadata, 
+            Column('id', Integer, Sequence('some_id_seq'), primary_key=True)
+        )
+
+Currently, when SQLAlchemy issues a single insert statement, to fulfill the contract of
+having the "last insert identifier" available, the sequence is executed independently
+beforehand and the new value is retrieved, to be used in the subsequent insert.  Note
+that when an :func:`~sqlalchemy.sql.expression.insert()` construct is executed using 
+"executemany" semantics, the sequence is not pre-executed and normal PG SERIAL behavior
+is used.
+
+Postgres 8.3 supports an ``INSERT...RETURNING`` syntax which SQLAlchemy supports 
+as well.  A future release of SQLA will use this feature by default in lieu of 
+sequence pre-execution in order to retrieve new primary key values, when available.
+
+INSERT/UPDATE...RETURNING
+-------------------------
+
+The dialect supports PG 8.3's ``INSERT..RETURNING`` and ``UPDATE..RETURNING`` syntaxes, 
+but must be explicitly enabled on a per-statement basis::
+
+    # INSERT..RETURNING
+    result = table.insert(postgres_returning=[table.c.col1, table.c.col2]).\\
+        values(name='foo')
+    print result.fetchall()
+    
+    # UPDATE..RETURNING
+    result = table.update(postgres_returning=[table.c.col1, table.c.col2]).\\
+        where(table.c.name=='foo').values(name='bar')
+    print result.fetchall()
+
+Indexes
+-------
+
+PostgreSQL supports partial indexes. To create them pass a postgres_where
 option to the Index constructor::
 
   Index('my_index', my_table.c.id, postgres_where=tbl.c.value > 10)
 
-PostgreSQL 8.2+ supports returning a result set from inserts and updates.
-To use this pass the column/expression list to the postgres_returning
-parameter when creating the queries::
-    
-  raises = tbl.update(empl.c.sales > 100, values=dict(salary=empl.c.salary * 1.1), 
-    postgres_returning=[empl.c.id, empl.c.salary]).execute().fetchall()
+Transactions
+------------
+
+The Postgres dialect fully supports SAVEPOINT and two-phase commit operations.
+
+
 """
 
-import re, random, warnings, string
+import decimal, random, re, string
 
-from sqlalchemy import sql, schema, exceptions, util
+from sqlalchemy import sql, schema, exc, util
 from sqlalchemy.engine import base, default
 from sqlalchemy.sql import compiler, expression
 from sqlalchemy.sql import operators as sql_operators
@@ -32,12 +102,20 @@ class PGInet(sqltypes.TypeEngine):
     def get_col_spec(self):
         return "INET"
 
+class PGCidr(sqltypes.TypeEngine):
+    def get_col_spec(self):
+        return "CIDR"
+
+class PGMacAddr(sqltypes.TypeEngine):
+    def get_col_spec(self):
+        return "MACADDR"
+
 class PGNumeric(sqltypes.Numeric):
     def get_col_spec(self):
         if not self.precision:
             return "NUMERIC"
         else:
-            return "NUMERIC(%(precision)s, %(length)s)" % {'precision': self.precision, 'length' : self.length}
+            return "NUMERIC(%(precision)s, %(scale)s)" % {'precision': self.precision, 'scale' : self.scale}
 
     def bind_processor(self, dialect):
         return None
@@ -47,12 +125,12 @@ class PGNumeric(sqltypes.Numeric):
             return None
         else:
             def process(value):
-                if isinstance(value, util.decimal_type):
+                if isinstance(value, decimal.Decimal):
                     return float(value)
                 else:
                     return value
             return process
-        
+
 class PGFloat(sqltypes.Float):
     def get_col_spec(self):
         if not self.precision:
@@ -89,17 +167,23 @@ class PGInterval(sqltypes.TypeEngine):
     def get_col_spec(self):
         return "INTERVAL"
 
-class PGText(sqltypes.TEXT):
+class PGText(sqltypes.Text):
     def get_col_spec(self):
         return "TEXT"
 
 class PGString(sqltypes.String):
     def get_col_spec(self):
-        return "VARCHAR(%(length)s)" % {'length' : self.length}
+        if self.length:
+            return "VARCHAR(%(length)d)" % {'length' : self.length}
+        else:
+            return "VARCHAR"
 
 class PGChar(sqltypes.CHAR):
     def get_col_spec(self):
-        return "CHAR(%(length)s)" % {'length' : self.length}
+        if self.length:
+            return "CHAR(%(length)d)" % {'length' : self.length}
+        else:
+            return "CHAR"
 
 class PGBinary(sqltypes.Binary):
     def get_col_spec(self):
@@ -109,25 +193,48 @@ class PGBoolean(sqltypes.Boolean):
     def get_col_spec(self):
         return "BOOLEAN"
 
-class PGArray(sqltypes.Concatenable, sqltypes.TypeEngine):
-    def __init__(self, item_type):
+class PGBit(sqltypes.TypeEngine):
+    def get_col_spec(self):
+        return "BIT"
+        
+class PGUuid(sqltypes.TypeEngine):
+    def get_col_spec(self):
+        return "UUID"
+    
+class PGArray(sqltypes.MutableType, sqltypes.Concatenable, sqltypes.TypeEngine):
+    def __init__(self, item_type, mutable=True):
         if isinstance(item_type, type):
             item_type = item_type()
         self.item_type = item_type
-        
-    def dialect_impl(self, dialect):
+        self.mutable = mutable
+
+    def copy_value(self, value):
+        if value is None:
+            return None
+        elif self.mutable:
+            return list(value)
+        else:
+            return value
+
+    def compare_values(self, x, y):
+        return x == y
+
+    def is_mutable(self):
+        return self.mutable
+
+    def dialect_impl(self, dialect, **kwargs):
         impl = self.__class__.__new__(self.__class__)
         impl.__dict__.update(self.__dict__)
         impl.item_type = self.item_type.dialect_impl(dialect)
         return impl
-        
+
     def bind_processor(self, dialect):
         item_proc = self.item_type.bind_processor(dialect)
         def process(value):
             if value is None:
                 return value
             def convert_item(item):
-                if isinstance(item, (list,tuple)):
+                if isinstance(item, (list, tuple)):
                     return [convert_item(child) for child in item]
                 else:
                     if item_proc:
@@ -136,9 +243,9 @@ class PGArray(sqltypes.Concatenable, sqltypes.TypeEngine):
                         return item
             return [convert_item(item) for item in value]
         return process
-        
+
     def result_processor(self, dialect):
-        item_proc = self.item_type.bind_processor(dialect)
+        item_proc = self.item_type.result_processor(dialect)
         def process(value):
             if value is None:
                 return value
@@ -166,7 +273,7 @@ colspecs = {
     sqltypes.String : PGString,
     sqltypes.Binary : PGBinary,
     sqltypes.Boolean : PGBoolean,
-    sqltypes.TEXT : PGText,
+    sqltypes.Text : PGText,
     sqltypes.CHAR: PGChar,
 }
 
@@ -176,11 +283,17 @@ ischema_names = {
     'smallint' : PGSmallInteger,
     'character varying' : PGString,
     'character' : PGChar,
+    '"char"' : PGChar,
+    'name': PGChar,
     'text' : PGText,
     'numeric' : PGNumeric,
     'float' : PGFloat,
     'real' : PGFloat,
     'inet': PGInet,
+    'cidr': PGCidr,
+    'uuid':PGUuid,
+    'bit':PGBit,
+    'macaddr': PGMacAddr,
     'double precision' : PGFloat,
     'timestamp' : PGDateTime,
     'timestamp with time zone' : PGDateTime,
@@ -194,86 +307,41 @@ ischema_names = {
     'interval':PGInterval,
 }
 
-def descriptor():
-    return {'name':'postgres',
-    'description':'PostGres',
-    'arguments':[
-        ('username',"Database Username",None),
-        ('password',"Database Password",None),
-        ('database',"Database Name",None),
-        ('host',"Hostname", None),
-    ]}
-
-SELECT_RE = re.compile(
-    r'\s*(?:SELECT|FETCH|(UPDATE|INSERT))',
+# TODO: filter out 'FOR UPDATE' statements
+SERVER_SIDE_CURSOR_RE = re.compile(
+    r'\s*SELECT',
     re.I | re.UNICODE)
 
-RETURNING_RE = re.compile(
-    'RETURNING',
-    re.I | re.UNICODE)
-
-# This finds if the RETURNING is not inside a quoted/commented values. Handles string literals,
-# quoted identifiers, dollar quotes, SQL comments and C style multiline comments. This does not
-# handle correctly nested C style quotes, lets hope no one does the following:
-# UPDATE tbl SET x=y /* foo /* bar */ RETURNING */
-RETURNING_QUOTED_RE = re.compile(
-    """\s*(?:UPDATE|INSERT)\s
-        (?: # handle quoted and commented tokens separately
-            [^'"$/-] # non quote/comment character
-            | -(?!-) # a dash that does not begin a comment
-            | /(?!\*) # a slash that does not begin a comment
-            | "(?:[^"]|"")*" # quoted literal
-            | '(?:[^']|'')*' # quoted string
-            | \$(?P<dquote>[^$]*)\$.*?\$(?P=dquote)\$ # dollar quotes
-            | --[^\\n]*(?=\\n) # SQL comment, leave out line ending as that counts as whitespace
-                            # for the returning token
-            | /\*([^*]|\*(?!/))*\*/ # C style comment, doesn't handle nesting
-        )*
-        \sRETURNING\s""", re.I | re.UNICODE | re.VERBOSE)
-
 class PGExecutionContext(default.DefaultExecutionContext):
-
-    def is_select(self):
-        m = SELECT_RE.match(self.statement)
-        return m and (not m.group(1) or (RETURNING_RE.search(self.statement)
-           and RETURNING_QUOTED_RE.match(self.statement)))
-        
     def create_cursor(self):
-        # executing a default or Sequence standalone creates an execution context without a statement.  
-        # so slightly hacky "if no statement assume we're server side" logic
-        self.__is_server_side = \
-            self.dialect.server_side_cursors and (self.statement is None or \
-            (SELECT_RE.match(self.statement) and not re.search(r'FOR UPDATE(?: NOWAIT)?\s*$', self.statement, re.I))
-        )
+        # TODO: coverage for server side cursors + select.for_update()
+        is_server_side = \
+            self.dialect.server_side_cursors and \
+            ((self.compiled and isinstance(self.compiled.statement, expression.Selectable) 
+                and not getattr(self.compiled.statement, 'for_update', False)) \
+            or \
+            (
+                (not self.compiled or isinstance(self.compiled.statement, expression._TextClause)) 
+                and self.statement and SERVER_SIDE_CURSOR_RE.match(self.statement))
+            )
 
-        if self.__is_server_side:
+        self.__is_server_side = is_server_side
+        if is_server_side:
             # use server-side cursors:
             # http://lists.initd.org/pipermail/psycopg/2007-January/005251.html
-            ident = "c" + hex(random.randint(0, 65535))[2:]
+            ident = "c_%s_%s" % (hex(id(self))[2:], hex(random.randint(0, 65535))[2:])
             return self._connection.connection.cursor(ident)
         else:
             return self._connection.connection.cursor()
-
+    
     def get_result_proxy(self):
         if self.__is_server_side:
             return base.BufferedRowResultProxy(self)
         else:
             return base.ResultProxy(self)
-    
-    def post_exec(self):
-        if self.compiled.isinsert and self.last_inserted_ids is None:
-            if not self.dialect.use_oids:
-                pass
-                # will raise invalid error when they go to get them
-            else:
-                table = self.compiled.statement.table
-                if self.cursor.lastrowid is not None and table is not None and len(table.primary_key):
-                    s = sql.select(table.primary_key, table.oid_column == self.cursor.lastrowid)
-                    row = self.connection.execute(s).fetchone()
-                self._last_inserted_ids = [v for v in row]
-        super(PGExecutionContext, self).post_exec()
-        
+
 class PGDialect(default.DefaultDialect):
+    name = 'postgres'
     supports_alter = True
     supports_unicode_statements = False
     max_identifier_length = 63
@@ -281,18 +349,19 @@ class PGDialect(default.DefaultDialect):
     supports_sane_multi_rowcount = False
     preexecute_pk_sequences = True
     supports_pk_autoincrement = False
-
-    def __init__(self, use_oids=False, server_side_cursors=False, **kwargs):
-        default.DefaultDialect.__init__(self, default_paramstyle='pyformat', **kwargs)
-        self.use_oids = use_oids
+    default_paramstyle = 'pyformat'
+    supports_default_values = True
+    supports_empty_insert = False
+    
+    def __init__(self, server_side_cursors=False, **kwargs):
+        default.DefaultDialect.__init__(self, **kwargs)
         self.server_side_cursors = server_side_cursors
-        self.paramstyle = 'pyformat'
-        
+
     def dbapi(cls):
         import psycopg2 as psycopg
         return psycopg
     dbapi = classmethod(dbapi)
-    
+
     def create_connect_args(self, url):
         opts = url.translate_connect_args(username='user')
         if 'port' in opts:
@@ -300,9 +369,6 @@ class PGDialect(default.DefaultDialect):
         opts.update(url.query)
         return ([], opts)
 
-    def create_execution_context(self, *args, **kwargs):
-        return PGExecutionContext(self, *args, **kwargs)
-
     def type_descriptor(self, typeobj):
         return sqltypes.adapt_type(typeobj, colspecs)
 
@@ -339,22 +405,16 @@ class PGDialect(default.DefaultDialect):
         return [row[0] for row in resultset]
 
     def get_default_schema_name(self, connection):
-        if not hasattr(self, '_default_schema_name'):
-            self._default_schema_name = connection.scalar("select current_schema()", None)
-        return self._default_schema_name
+        return connection.scalar("select current_schema()", None)
+    get_default_schema_name = base.connection_memoize(
+        ('dialect', 'default_schema_name'))(get_default_schema_name)
 
     def last_inserted_ids(self):
         if self.context.last_inserted_ids is None:
-            raise exceptions.InvalidRequestError("no INSERT executed, or can't use cursor.lastrowid without Postgres OIDs enabled")
+            raise exc.InvalidRequestError("no INSERT executed, or can't use cursor.lastrowid without Postgres OIDs enabled")
         else:
             return self.context.last_inserted_ids
 
-    def oid_column_name(self, column):
-        if self.use_oids:
-            return "oid"
-        else:
-            return None
-
     def has_table(self, connection, table_name, schema=None):
         # seems like case gets folded in pg_class...
         if schema is None:
@@ -371,16 +431,16 @@ class PGDialect(default.DefaultDialect):
         if isinstance(e, self.dbapi.OperationalError):
             return 'closed the connection' in str(e) or 'connection not open' in str(e)
         elif isinstance(e, self.dbapi.InterfaceError):
-            return 'connection already closed' in str(e)
+            return 'connection already closed' in str(e) or 'cursor already closed' in str(e)
         elif isinstance(e, self.dbapi.ProgrammingError):
             # yes, it really says "losed", not "closed"
             return "losed the connection unexpectedly" in str(e)
         else:
             return False
-        
+
     def table_names(self, connection, schema):
         s = """
-        SELECT relname 
+        SELECT relname
         FROM pg_class c
         WHERE relkind = 'r'
           AND '%(schema)s' = (select nspname from pg_namespace n where n.oid = c.relnamespace)
@@ -391,18 +451,19 @@ class PGDialect(default.DefaultDialect):
         v = connection.execute("select version()").scalar()
         m = re.match('PostgreSQL (\d+)\.(\d+)\.(\d+)', v)
         if not m:
-            raise exceptions.AssertionError("Could not determine version from string '%s'" % v)
+            raise AssertionError("Could not determine version from string '%s'" % v)
         return tuple([int(x) for x in m.group(1, 2, 3)])
 
     def reflecttable(self, connection, table, include_columns):
         preparer = self.identifier_preparer
         if table.schema is not None:
             schema_where_clause = "n.nspname = :schema"
+            schemaname = table.schema
+            if isinstance(schemaname, str):
+                schemaname = schemaname.decode(self.encoding)
         else:
             schema_where_clause = "pg_catalog.pg_table_is_visible(c.oid)"
-
-        ## information schema in pg suffers from too many permissions' restrictions
-        ## let us find out at the pg way what is needed...
+            schemaname = None
 
         SQL_COLS = """
             SELECT a.attname,
@@ -423,19 +484,21 @@ class PGDialect(default.DefaultDialect):
         """ % schema_where_clause
 
         s = sql.text(SQL_COLS, bindparams=[sql.bindparam('table_name', type_=sqltypes.Unicode), sql.bindparam('schema', type_=sqltypes.Unicode)], typemap={'attname':sqltypes.Unicode, 'default':sqltypes.Unicode})
-        c = connection.execute(s, table_name=table.name,
-                                  schema=table.schema)
+        tablename = table.name
+        if isinstance(tablename, str):
+            tablename = tablename.decode(self.encoding)
+        c = connection.execute(s, table_name=tablename, schema=schemaname)
         rows = c.fetchall()
 
         if not rows:
-            raise exceptions.NoSuchTableError(table.name)
+            raise exc.NoSuchTableError(table.name)
 
         domains = self._load_domains(connection)
-        
+
         for name, format_type, default, notnull, attnum, table_oid in rows:
             if include_columns and name not in include_columns:
                 continue
-            
+
             ## strip (30) from character varying(30)
             attype = re.search('([^\([]+)', format_type).group(1)
             nullable = not notnull
@@ -474,6 +537,7 @@ class PGDialect(default.DefaultDialect):
             elif attype == 'timestamp without time zone':
                 kwargs['timezone'] = False
 
+            coltype = None
             if attype in ischema_names:
                 coltype = ischema_names[attype]
             else:
@@ -487,18 +551,17 @@ class PGDialect(default.DefaultDialect):
                             # It can, however, override the default value, but can't set it to null.
                             default = domain['default']
                         coltype = ischema_names[domain['attype']]
-                else:
-                    coltype=None
 
             if coltype:
                 coltype = coltype(*args, **kwargs)
                 if is_array:
                     coltype = PGArray(coltype)
             else:
-                warnings.warn(RuntimeWarning("Did not recognize type '%s' of column '%s'" % (attype, name)))
+                util.warn("Did not recognize type '%s' of column '%s'" %
+                          (attype, name))
                 coltype = sqltypes.NULLTYPE
 
-            colargs= []
+            colargs = []
             if default is not None:
                 match = re.search(r"""(nextval\(')([^']+)('.*$)""", default)
                 if match is not None:
@@ -508,7 +571,7 @@ class PGDialect(default.DefaultDialect):
                         # unconditionally quote the schema name.  this could
                         # later be enhanced to obey quoting rules / "quote schema"
                         default = match.group(1) + ('"%s"' % sch) + '.' + match.group(2) + match.group(3)
-                colargs.append(schema.PassiveDefault(sql.text(default)))
+                colargs.append(schema.DefaultClause(sql.text(default)))
             table.append_column(schema.Column(name, coltype, nullable=nullable, *colargs))
 
 
@@ -525,10 +588,11 @@ class PGDialect(default.DefaultDialect):
         c = connection.execute(t, table=table_oid)
         for row in c.fetchall():
             pk = row[0]
-            col = table.c[pk]
-            table.primary_key.add(col)
-            if col.default is None:
-                col.autoincrement=False
+            if pk in table.c:
+                col = table.c[pk]
+                table.primary_key.add(col)
+                if col.default is None:
+                    col.autoincrement = False
 
         # Foreign keys
         FK_SQL = """
@@ -547,7 +611,7 @@ class PGDialect(default.DefaultDialect):
             if referred_schema:
                 referred_schema = preparer._unquote_identifier(referred_schema)
             elif table.schema is not None and table.schema == self.get_default_schema_name(connection):
-                # no schema (i.e. its the default schema), and the table we're 
+                # no schema (i.e. its the default schema), and the table we're
                 # reflecting has the default schema explicit, then use that.
                 # i.e. try to use the user's conventions
                 referred_schema = table.schema
@@ -565,8 +629,47 @@ class PGDialect(default.DefaultDialect):
                 for column in referred_columns:
                     refspec.append(".".join([referred_table, column]))
 
-            table.append_constraint(schema.ForeignKeyConstraint(constrained_columns, refspec, conname))
-                
+            table.append_constraint(schema.ForeignKeyConstraint(constrained_columns, refspec, conname, link_to_name=True))
+
+        # Indexes 
+        IDX_SQL = """
+          SELECT c.relname, i.indisunique, i.indexprs, i.indpred,
+            a.attname
+          FROM pg_index i, pg_class c, pg_attribute a
+          WHERE i.indrelid = :table AND i.indexrelid = c.oid
+            AND a.attrelid = i.indexrelid AND i.indisprimary = 'f'
+          ORDER BY c.relname, a.attnum
+        """
+        t = sql.text(IDX_SQL, typemap={'attname':sqltypes.Unicode})
+        c = connection.execute(t, table=table_oid)
+        indexes = {}
+        sv_idx_name = None
+        for row in c.fetchall():
+            idx_name, unique, expr, prd, col = row
+
+            if expr:
+                if not idx_name == sv_idx_name:
+                    util.warn(
+                      "Skipped unsupported reflection of expression-based index %s"
+                      % idx_name)
+                sv_idx_name = idx_name
+                continue
+            if prd and not idx_name == sv_idx_name:
+                util.warn(
+                   "Predicate of partial index %s ignored during reflection"
+                   % idx_name)
+                sv_idx_name = idx_name
+
+            if not indexes.has_key(idx_name):
+                indexes[idx_name] = [unique, []]
+            indexes[idx_name][1].append(col)
+
+        for name, (unique, columns) in indexes.items():
+            schema.Index(name, *[table.columns[c] for c in columns], 
+                         **dict(unique=unique))
+ 
+
+
     def _load_domains(self, connection):
         ## Load data types for domains:
         SQL_DOMAINS = """
@@ -590,7 +693,7 @@ class PGDialect(default.DefaultDialect):
             ## strip (30) from character varying(30)
             attype = re.search('([^\(]+)', domain['attype']).group(1)
             if domain['visible']:
-                # 'visible' just means whether or not the domain is in a 
+                # 'visible' just means whether or not the domain is in a
                 # schema that's on the search path -- or not overriden by
                 # a schema with higher presedence. If it's not visible,
                 # it will be prefixed with the schema-name when it's used.
@@ -601,14 +704,23 @@ class PGDialect(default.DefaultDialect):
             domains[name] = {'attype':attype, 'nullable': domain['nullable'], 'default': domain['default']}
 
         return domains
-        
-        
-        
+
+
 class PGCompiler(compiler.DefaultCompiler):
     operators = compiler.DefaultCompiler.operators.copy()
     operators.update(
         {
-            sql_operators.mod : '%%'
+            sql_operators.mod : '%%',
+            sql_operators.ilike_op: lambda x, y, escape=None: '%s ILIKE %s' % (x, y) + (escape and ' ESCAPE \'%s\'' % escape or ''),
+            sql_operators.notilike_op: lambda x, y, escape=None: '%s NOT ILIKE %s' % (x, y) + (escape and ' ESCAPE \'%s\'' % escape or ''),
+            sql_operators.match_op: lambda x, y: '%s @@ to_tsquery(%s)' % (x, y),
+        }
+    )
+
+    functions = compiler.DefaultCompiler.functions.copy()
+    functions.update (
+        {
+            'TIMESTAMP':util.deprecated(message="Use a literal string 'timestamp <value>' instead")(lambda x:'TIMESTAMP %s' % x),
         }
     )
 
@@ -617,7 +729,12 @@ class PGCompiler(compiler.DefaultCompiler):
             return None
         else:
             return "nextval('%s')" % self.preparer.format_sequence(seq)
-        
+
+    def post_process_text(self, text):
+        if '%%' in text:
+            util.warn("The SQLAlchemy psycopg2 dialect now automatically escapes '%' in text() expressions to '%%'.")
+        return text.replace('%', '%%')
+
     def limit_clause(self, select):
         text = ""
         if select._limit is not None:
@@ -656,7 +773,7 @@ class PGCompiler(compiler.DefaultCompiler):
                         yield co
                 else:
                     yield c
-        columns = [self.process(c) for c in flatten_columnlist(returning_cols)]
+        columns = [self.process(c, within_columns_clause=True) for c in flatten_columnlist(returning_cols)]
         text += ' RETURNING ' + string.join(columns, ', ')
         return text
 
@@ -674,6 +791,12 @@ class PGCompiler(compiler.DefaultCompiler):
         else:
             return text
 
+    def visit_extract(self, extract, **kwargs):
+        field = self.extract_map.get(extract.field, extract.field)
+        return "EXTRACT(%s FROM %s::timestamp)" % (
+            field, self.process(extract.expr))
+
+
 class PGSchemaGenerator(compiler.SchemaGenerator):
     def get_column_specification(self, column, **kwargs):
         colspec = self.preparer.format_column(column)
@@ -696,14 +819,14 @@ class PGSchemaGenerator(compiler.SchemaGenerator):
         if not sequence.optional and (not self.checkfirst or not self.dialect.has_sequence(self.connection, sequence.name)):
             self.append("CREATE SEQUENCE %s" % self.preparer.format_sequence(sequence))
             self.execute()
-    
+
     def visit_index(self, index):
         preparer = self.preparer
         self.append("CREATE ")
         if index.unique:
             self.append("UNIQUE ")
         self.append("INDEX %s ON %s (%s)" \
-                    % (preparer.format_index(index),
+                    % (preparer.quote(self._validate_identifier(index.name, True), index.quote),
                        preparer.format_table(index.table),
                        string.join([preparer.format_column(c) for c in index.columns], ', ')))
         whereclause = index.kwargs.get('postgres_where', None)
@@ -722,11 +845,17 @@ class PGSchemaDropper(compiler.SchemaDropper):
             self.execute()
 
 class PGDefaultRunner(base.DefaultRunner):
+    def __init__(self, context):
+        base.DefaultRunner.__init__(self, context)
+        # craete cursor which won't conflict with a server-side cursor
+        self.cursor = context._connection.connection.cursor()
+    
     def get_column_default(self, column, isinsert=True):
         if column.primary_key:
             # pre-execute passive defaults on primary keys
-            if isinstance(column.default, schema.PassiveDefault):
-                return self.execute_string("select %s" % column.default.arg)
+            if (isinstance(column.server_default, schema.DefaultClause) and
+                column.server_default.arg is not None):
+                return self.execute_string("select %s" % column.server_default.arg)
             elif (isinstance(column.type, sqltypes.Integer) and column.autoincrement) and (column.default is None or (isinstance(column.default, schema.Sequence) and column.default.optional)):
                 sch = column.table.schema
                 # TODO: this has to build into the Sequence object so we can get the quoting
@@ -741,7 +870,7 @@ class PGDefaultRunner(base.DefaultRunner):
 
     def visit_sequence(self, seq):
         if not seq.optional:
-            return self.execute_string(("select nextval('%s')" % self.dialect.identifier_preparer.format_sequence(seq)).encode(self.dialect.encoding))
+            return self.execute_string(("select nextval('%s')" % self.dialect.identifier_preparer.format_sequence(seq)))
         else:
             return None
 
@@ -757,3 +886,4 @@ dialect.schemagenerator = PGSchemaGenerator
 dialect.schemadropper = PGSchemaDropper
 dialect.preparer = PGIdentifierPreparer
 dialect.defaultrunner = PGDefaultRunner
+dialect.execution_ctx_cls = PGExecutionContext
