@@ -1,5 +1,5 @@
 # interfaces.py
-# Copyright (C) 2005, 2006, 2007, 2008 Michael Bayer mike_mp@zzzcomputing.com
+# Copyright (C) 2005, 2006, 2007, 2008, 2009 Michael Bayer mike_mp@zzzcomputing.com
 #
 # This module is part of SQLAlchemy and is released under
 # the MIT License: http://www.opensource.org/licenses/mit-license.php
@@ -8,7 +8,7 @@
 
 Semi-private module containing various base classes used throughout the ORM.
 
-Defines the extension classes :class:`MapperExtension`, 
+Defines the extension classes :class:`MapperExtension`,
 :class:`SessionExtension`, and :class:`AttributeExtension` as
 well as other user-subclassable extension objects.
 
@@ -167,7 +167,7 @@ class MapperExtension(object):
         ``__new__``, and after initial attribute population has
         occurred.
 
-        This typicically occurs when the instance is created based on
+        This typically occurs when the instance is created based on
         incoming result rows, and is only called once for that
         instance's lifetime.
 
@@ -325,7 +325,7 @@ class SessionExtension(object):
         `query_context` was the query context object.
         `result` is the result object returned from the bulk operation.
         """
-    
+
 class MapperProperty(object):
     """Manage the relationship of a ``Mapper`` to a single class
     attribute, as well as that attribute as it appears on individual
@@ -359,7 +359,7 @@ class MapperProperty(object):
 
         Callables are of the following form::
 
-            def new_execute(state, row, **flags):
+            def new_execute(state, dict_, row, **flags):
                 # process incoming instance state and given row.  the instance is
                 # "new" and was just created upon receipt of this row.
                 # flags is a dictionary containing at least the following
@@ -368,7 +368,7 @@ class MapperProperty(object):
                 #           result of reading this row
                 #   instancekey - identity key of the instance
 
-            def existing_execute(state, row, **flags):
+            def existing_execute(state, dict_, row, **flags):
                 # process incoming instance state and given row.  the instance is
                 # "existing" and was created based on a previous row.
 
@@ -392,34 +392,58 @@ class MapperProperty(object):
     def set_parent(self, parent):
         self.parent = parent
 
-    def init(self, key, parent):
-        """Called after all mappers are compiled to assemble
-        relationships between mappers, establish instrumented class
-        attributes.
-        """
+    def instrument_class(self, mapper):
+        raise NotImplementedError()
 
-        self.key = key
-        self._compiled = True
+    _compile_started = False
+    _compile_finished = False
+
+    def init(self):
+        """Called after all mappers are created to assemble
+        relationships between mappers and perform other post-mapper-creation
+        initialization steps.
+
+        """
+        self._compile_started = True
         self.do_init()
+        self._compile_finished = True
 
     def do_init(self):
-        """Perform subclass-specific initialization steps.
+        """Perform subclass-specific initialization post-mapper-creation steps.
 
         This is a *template* method called by the
-        ``MapperProperty`` object's init() method."""
+        ``MapperProperty`` object's init() method.
 
+        """
+        pass
+
+    def post_instrument_class(self, mapper):
+        """Perform instrumentation adjustments that need to occur
+        after init() has completed.
+
+        """
         pass
 
     def register_dependencies(self, *args, **kwargs):
         """Called by the ``Mapper`` in response to the UnitOfWork
         calling the ``Mapper``'s register_dependencies operation.
-        Should register with the UnitOfWork all inter-mapper
-        dependencies as well as dependency processors (see UOW docs
-        for more details).
+        Establishes a topological dependency between two mappers
+        which will affect the order in which mappers persist data.
+        
         """
 
         pass
 
+    def register_processors(self, *args, **kwargs):
+        """Called by the ``Mapper`` in response to the UnitOfWork
+        calling the ``Mapper``'s register_processors operation.
+        Establishes a processor object between two mappers which
+        will link data and state between parent/child objects.
+        
+        """
+
+        pass
+        
     def is_primary(self):
         """Return True if this ``MapperProperty``'s mapper is the
         primary mapper for its class.
@@ -468,10 +492,10 @@ class PropComparator(expression.ColumnOperators):
     def adapted(self, adapter):
         """Return a copy of this PropComparator which will use the given adaption function
         on the local side of generated expressions.
-        
+
         """
         return self.__class__(self.prop, self.mapper, adapter)
-        
+
     @staticmethod
     def any_op(a, b, **kwargs):
         return a.any(b, **kwargs)
@@ -571,8 +595,10 @@ class StrategizedProperty(MapperProperty):
     def do_init(self):
         self.__all_strategies = {}
         self.strategy = self.__init_strategy(self.strategy_class)
+
+    def post_instrument_class(self, mapper):
         if self.is_primary():
-            self.strategy.init_class_attribute()
+            self.strategy.init_class_attribute(mapper)
 
 def build_path(entity, key, prev=None):
     if prev:
@@ -666,6 +692,27 @@ class PropertyOption(MapperOption):
             else:
                 return None
 
+    def __getstate__(self):
+        d = self.__dict__.copy()
+        d['key'] = ret = []
+        for token in util.to_list(self.key):
+            if isinstance(token, PropComparator):
+                ret.append((token.mapper.class_, token.key))
+            else:
+                ret.append(token)
+        return d
+
+    def __setstate__(self, state):
+        ret = []
+        for key in state['key']:
+            if isinstance(key, tuple):
+                cls, propkey = key
+                ret.append(getattr(cls, propkey))
+            else:
+                ret.append(key)
+        state['key'] = tuple(ret)
+        self.__dict__ = state
+
     def __get_paths(self, query, raiseerr):
         path = None
         entity = None
@@ -722,35 +769,35 @@ class PropertyOption(MapperOption):
 
 class AttributeExtension(object):
     """An event handler for individual attribute change events.
-    
-    AttributeExtension is assembled within the descriptors associated 
-    with a mapped class. 
-    
+
+    AttributeExtension is assembled within the descriptors associated
+    with a mapped class.
+
     """
 
     def append(self, state, value, initiator):
         """Receive a collection append event.
-        
+
         The returned value will be used as the actual value to be
         appended.
-        
+
         """
         return value
 
     def remove(self, state, value, initiator):
         """Receive a remove event.
-        
+
         No return value is defined.
-        
+
         """
         pass
 
     def set(self, state, value, oldvalue, initiator):
         """Receive a set event.
-        
+
         The returned value will be used as the actual value to be
         set.
-        
+
         """
         return value
 
@@ -808,7 +855,7 @@ class LoaderStrategy(object):
     def init(self):
         raise NotImplementedError("LoaderStrategy")
 
-    def init_class_attribute(self):
+    def init_class_attribute(self, mapper):
         pass
 
     def setup_query(self, context, entity, path, adapter, **kwargs):
@@ -839,7 +886,12 @@ class LoaderStrategy(object):
             return fn
 
 class InstrumentationManager(object):
-    """User-defined class instrumentation extension."""
+    """User-defined class instrumentation extension.
+    
+    The API for this class should be considered as semi-stable,
+    and may change slightly with new releases.
+    
+    """
 
     # r4361 added a mandatory (cls) constructor to this interface.
     # given that, perhaps class_ should be dropped from all of these
@@ -862,6 +914,9 @@ class InstrumentationManager(object):
     def instrument_attribute(self, class_, key, inst):
         pass
 
+    def post_configure_attribute(self, class_, key, inst):
+        pass
+
     def install_descriptor(self, class_, key, inst):
         setattr(class_, key, inst)
 
@@ -894,3 +949,7 @@ class InstrumentationManager(object):
 
     def state_getter(self, class_):
         return lambda instance: getattr(instance, '_default_state')
+
+    def dict_getter(self, class_):
+        return lambda inst: self.get_instance_dict(class_, inst)
+        
\ No newline at end of file
