@@ -1,5 +1,5 @@
 # types.py
-# Copyright (C) 2005, 2006, 2007, 2008 Michael Bayer mike_mp@zzzcomputing.com
+# Copyright (C) 2005, 2006, 2007, 2008, 2009 Michael Bayer mike_mp@zzzcomputing.com
 #
 # This module is part of SQLAlchemy and is released under
 # the MIT License: http://www.opensource.org/licenses/mit-license.php
@@ -24,7 +24,7 @@ __all__ = [ 'TypeEngine', 'TypeDecorator', 'AbstractType',
 import inspect
 import datetime as dt
 from decimal import Decimal as _python_Decimal
-
+import weakref
 from sqlalchemy import exc
 from sqlalchemy.util import pickle
 import sqlalchemy.util as util
@@ -125,14 +125,14 @@ class TypeEngine(AbstractType):
         try:
             return self._impl_dict[dialect]
         except AttributeError:
-            self._impl_dict = {}
+            self._impl_dict = weakref.WeakKeyDictionary()   # will be optimized in 0.6
             return self._impl_dict.setdefault(dialect, dialect.type_descriptor(self))
         except KeyError:
             return self._impl_dict.setdefault(dialect, dialect.type_descriptor(self))
 
     def __getstate__(self):
         d = self.__dict__.copy()
-        d['_impl_dict'] = {}
+        d.pop('_impl_dict', None)
         return d
 
     def get_col_spec(self):
@@ -223,7 +223,7 @@ class TypeDecorator(AbstractType):
         try:
             return self._impl_dict[dialect]
         except AttributeError:
-            self._impl_dict = {}
+            self._impl_dict = weakref.WeakKeyDictionary()   # will be optimized in 0.6
         except KeyError:
             pass
 
@@ -485,8 +485,6 @@ class Text(String):
     params (and the reverse for result sets.)
 
     """
-    def dialect_impl(self, dialect, **kwargs):
-        return TypeEngine.dialect_impl(self, dialect, **kwargs)
 
 class Unicode(String):
     """A variable length Unicode string.
